;; Generated package description from dape.el  -*- no-byte-compile: t -*-
(define-package "dape" "0.2.2.0.20231220.182354" "Debug Adapter Protocol for Emacs" '((emacs "29.1")) :commit "28f43ea21ce73e4480168321c1593eedc20aa5ce" :maintainer '("Daniel Pettersson" . "daniel@dpettersson.net") :url "https://github.com/svaante/dape")
